import os
from .util import gridify, import_torch, _create_folder
import cv2, numpy as np

torch = import_torch()


@torch.no_grad()
def show_video(
    tensor: torch.Tensor,
    fps: int = 30,
    columns: int | None = None,
    max_width: int | None = None,
    padding: int = 3,
    pad_value: float = 0.0,
) -> None:
    """
    Shows tensor as a video. Accepts both **(T,H,W)**, **(T,3,H,W)** and **(\\*,T,3,H,W)** tensors.
    Tensor should be a float tensor with values in [0,1]. Clips the values otherwise.

    NOT IMPLEMENTED YET

    Args:
        tensor : (T,H,W) or (T,3,H,W) or (\\*,T,3,H,W) float tensor
        columns : number of columns to use for the grid of videos (default 8 or less)
        fps : fps of the video (default 30)
        out_size : Height of output video (height adapts to not deform videos) (default 800)
    """
    return NotImplementedError("showVideo not implemented yet, use saveVideo instead")


@torch.no_grad()
def save_video(
    tensor: torch.Tensor,
    folder: str,
    name: str = "videotensor",
    fps: int = 30,
    columns: int | None = None,
    max_width: int | None = None,
    padding: int = 3,
    pad_value: float = 0.0,
    create_folder: bool = True,
) -> None:
    """
    Shows tensor as a video. Accepts both **(T,H,W)**, **(T,3,H,W)** and **(\\*,T,3,H,W)** tensors.
    Tensor should be a float tensor with values in [0,1]. Clips the values otherwise.

    Args:
        tensor : (T,H,W) or (T,3,H,W) or (\\*,T,3,H,W) float tensor
        folder : path to save the video
        name : name of the video
        fps : fps of the video (default 30)
        columns : number of columns to use for the grid of videos (default 8 or less)
        max_width : maximum width of the image
        padding : number of pixels between images in the grid
        pad_value : inter-padding value for the grid of images
    """
    _create_folder(folder, create_folder)

    tensor = _format_video(
        tensor, columns=columns, fps=fps, max_width=max_width, padding=padding, pad_value=pad_value
    )  # (T,3,H',W') ready to save
    T, C, H, W = tensor.shape
    output_file = os.path.join(folder, f"{name}.mp4")
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    video = cv2.VideoWriter(output_file, fourcc, fps, (W, H))

    to_save = (255 * tensor.permute(0, 2, 3, 1).cpu().numpy()).astype(np.uint8)

    for t in range(T):
        frame = to_save[t]
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        video.write(frame)

    video.release()


def _format_video(tensor, columns=None, fps=30, max_width=None, padding=3, pad_value=0.0):
    """
    Formats tensor as a video. Accepts both (T,H,W), (T,3,H,W) and (\\*,T,3,H,W) float tensors.
    Assumes that the tensor value are in [0,1], clips them otherwise.

    Args:
        tensor : (T,H,W) or (T,3,H,W) or (\\*,T,3,H,W) float tensor
        columns : number of columns to use for the grid of videos (default 8 or less)
        fps : fps of the video (default 30)
        max_width : maximum width of the image
        padding : number of pixels between images in the grid
        pad_value : inter-padding value for the grid of images
    """
    tensor = tensor.detach().cpu()
    extra_params = dict(columns=columns, fps=fps, max_width=max_width, pad_value=pad_value, padding=padding)

    if len(tensor.shape) == 3:
        # add channel dimension
        tensor = tensor[:, None, :, :].expand(-1, 3, -1, -1)  # (T,3,H,W)
        return _format_video(tensor, **extra_params)
    elif len(tensor.shape) == 4:
        if tensor.shape[1] == 1:
            print("Assuming gray-scale video")
            tensor = tensor.expand(-1, 3, -1, -1)  # (T,3,H,W)
        assert tensor.shape[1] == 3, f"Tensor shape should be (T,H,W), (T,3,H,W) or (*,T,3,H,W) !"
        # A single video
        return tensor
    elif len(tensor.shape) == 5:
        tensor = gridify(tensor, max_width=max_width, columns=columns, padding=padding, pad_value=pad_value)
        return _format_video(tensor, **extra_params)
    elif len(tensor.shape) > 5:
        tensor = tensor.reshape((-1, *tensor.shape[-4:]))
        return _format_video(tensor, **extra_params)
    else:
        raise ValueError(f"Tensor shape should be (T,H,W), (T,3,H,W) or (*,T,3,H,W), but got : {tensor.shape} !")



import os
import matplotlib.pyplot as plt
from .imports import import_torch, import_torchvision
from .util import gridify, _create_folder, _format_image

torch = import_torch()
torchvision = import_torchvision()


@torch.no_grad()
def show_image(
    tensor: torch.Tensor,
    columns: int | None = None,
    rescale: bool = False,
    clamp_range: tuple[float] = (0.0, 1.0),
    colorbar: bool = False,
    max_width: int | None = None,
    padding: int = 3,
    pad_value: float = 0.0,
) -> None:
    """
    Shows tensor of shape **(\\*,C,H,W)** as an image using pyplot.
    Any extra dimensions are treated as batch dimensions, and displayed in a grid.
    By default, the image is simply clipped to the 0-1 range. Use rescale and clamp_range to modify this behaviour.

    Args:
        tensor : (H,W) or (C,H,W) or (\\*,C,H,W) tensor to display
        columns : number of columns to use for the grid of images (default 8 or less)
        rescale : whether to rescale the images to 0-1 range. Uses min-max scaling.
        clamp_range : (min,max), tuple of values to clamp the images to, before rescaling. Use None to disable clamping
        colorbar : whether to add a colorbar to the image, only works for grayscale images (default False)
        max_width : maximum width of the image
        padding : number of pixels between images in the grid
        pad_value : inter-padding value for the grid of images
    """
    tensor = _format_image(
        tensor,
        columns=columns,
        rescale=rescale,
        clamp_range=clamp_range,
        max_width=max_width,
        padding=padding,
        pad_value=pad_value,
    )  # (C,H',W') ready to show
    C, H, W = tensor.shape
    plt.figure(figsize=(W / 50, H / 50), dpi=50)
    plt.subplots_adjust(left=0, right=1, bottom=0, top=1)  # Remove margins

    plt.imshow(tensor.permute((1, 2, 0)))
    plt.axis("off")
    if tensor.shape[0] == 1 and colorbar:
        plt.colorbar()
    plt.show()


@torch.no_grad()
def save_image(
    tensor: torch.Tensor,
    folder: str,
    name: str = "imagetensor",
    columns: int | None = None,
    rescale: bool = False,
    clamp_range: tuple[float] = (0.0, 1.0),
    colorbar: bool = False,
    max_width: int | None = None,
    padding: int = 3,
    pad_value: float = 0.0,
    create_folder: bool = True,
) -> None:
    """
    Saves tensor of shape **(\\*,C,H,W)** as an image using pyplot.
    Any extra dimensions are treated as batch dimensions, and displayed in a grid.
    By default, the image is simply clipped to the 0-1 range. Use rescale and clamp_range to modify this behaviour.

    Args:
        tensor : (H,W) or (C,H,W) or (\\*,C,H,W) tensor to display
        folder : relative path of folder where to save the image
        name : name of the image (do not include extension)
        columns : number of columns to use for the grid of images (default 8 or less)
        rescale : whether to rescale the images to 0-1 range
        clamp_range : (min,max), tuple of values to clamp the images to, before rescaling. Use None to disable clamping
        colorbar : whether to add a colorbar to the image, only works for grayscale images (default False)
        max_width : maximum width of the image
        padding : number of pixels between images in the grid
        pad_value : inter-padding value for the grid of images
        create_folder : whether to create the folder if it does not exist (default True)
    """
    _create_folder(folder, create_folder)

    tensor = _format_image(
        tensor,
        columns=columns,
        rescale=rescale,
        clamp_range=clamp_range,
        max_width=max_width,
        padding=padding,
        pad_value=pad_value,
    )  # (C,H',W') ready to save
    H, W = tensor.shape[1], tensor.shape[2]
    plt.figure(figsize=(W / 50, H / 50), dpi=50)
    plt.subplots_adjust(left=0, right=1, bottom=0, top=1)  # Remove margins

    plt.imshow(tensor.permute((1, 2, 0)), extent=[0, W, 0, H])
    plt.axis("off")
    if tensor.shape[0] == 1 and colorbar:
        plt.colorbar()

    plt.savefig(os.path.join(folder, f"{name}.png"), dpi=50, bbox_inches="tight", pad_inches=0)
    plt.close()



import sys, pathlib, os

sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())

from src.showtens.util import import_torch, import_torchvision

# None implemented yet
from src.showtens import show_image, save_image

curpath = pathlib.Path(__file__).parent


def test_view():
    """
    Unfortunately I see no way to write automatic unit tests for this function.
    """
    from PIL import Image

    torch = import_torch()
    transf = import_torchvision().transforms
    image = Image.open(os.path.join(curpath, "test_folder", "whale.png"))

    whale = transf.ToTensor()(image)  # (4,H,W);
    randAug = transf.RandomResizedCrop(whale.shape[-2:])
    whale_tile = torch.stack([randAug(whale) for _ in range(7)], dim=0)  # (9,3,H,W)

    alpha_noise = torch.rand(7, 1, *whale.shape[-2:])
    whale_tile[:, 3:, :, :] = whale_tile[:, 3:, :, :] * (alpha_noise)  # (9,4,H,W)
    num, _, H, W = whale_tile.shape
    print("first, expected shape : ", (H * 2, W * 4))
    show_image(whale_tile, columns=4, colorbar=False, max_width=None, padding=0, pad_value=1.0)
    print("Second, expected max_width : 500, figure shape : ", (H * 2, W * 4))
    show_image(whale_tile[:, 2:3], columns=None, colorbar=True, max_width=500, padding=3, pad_value=0.0)


def test_save():
    """
    Unfortunately I see no way to write automatic unit tests for this function.
    """
    from PIL import Image

    torch = import_torch()
    transf = import_torchvision().transforms
    image = Image.open(os.path.join(curpath, "test_folder", "whale.png"))

    whale = transf.ToTensor()(image)  # (4,H,W);
    randAug = transf.RandomResizedCrop(whale.shape[-2:])
    whale_tile = torch.stack([randAug(whale) for _ in range(9)], dim=0)  # (9,4,H,W)

    _, _, H, W = whale_tile.shape

    print("Expected shape : ", (H * 2, W * 5))
    save_image(
        whale_tile,
        folder="test_folder",
        name="whale_list",
        columns=5,
        colorbar=False,
        max_width=None,
        padding=0,
        pad_value=1.0,
    )
    save_image(
        whale_tile[:, 2:3],
        folder="test_folder",
        name="dawhales",
        columns=None,
        colorbar=True,
        max_width=500,
        padding=3,
        pad_value=0.0,
    )


# # test_save()
# test_view()
# test_save()



import sys, pathlib, os

sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())

from src.showtens.util import import_torch, import_torchvision

# None implemented yet
from src.showtens.videos import save_video

curpath = pathlib.Path(__file__).parent


def test_video():
    """
    Unfortunately I see no way to write automatic unit tests for this function.
    """
    read_video = import_torchvision().io.read_video
    transf = import_torchvision().transforms
    torch = import_torch()

    randAug = transf.RandomResizedCrop(224)

    video, _, _ = read_video(
        os.path.join(curpath, "test_folder/test_vid.mp4"), output_format="TCHW", pts_unit="sec"
    )

    video = video.to(torch.float32) / 255.0
    video = video[::2]

    video_tile = torch.stack([randAug(video) for _ in range(3)], dim=0)  # (10,T,3,H,W)
    save_video(
        video_tile,
        folder="test_folder",
        name="see_same",
        fps=30,
        columns=2,
        max_width=None,
        padding=3,
        pad_value=1.0,
        create_folder=True,
    )



def import_torch():
    """
    Lazily import torch, giving helpful error message if not installed.
    """
    try:
        import torch

        return torch
    except ImportError:
        raise ImportError(
            "PyTorch is not installed. ShowTens requires PyTorch. "
            "Please install it with one of the following commands:\n\n"
            "- For CPU only: pip install torch\n"
            "- For CUDA support: Visit https://pytorch.org for installation instructions "
            "specific to your system and CUDA version."
        )


def import_torchvision():
    """
    Lazily import torchvision.transforms, giving helpful error message if not installed.
    """
    try:
        import torchvision

        return torchvision
    except ImportError:
        raise ImportError(
            "torchvision is not installed. ShowTens requires torchvision. "
            "Please install it with the following command:\n\n"
            "pip install torchvision"
        )


# torch = import_torch()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'showtens'
copyright = '2024, Vassilis Papadopoulos'
author = 'Vassilis Papadopoulos'
release = '0.3'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode'
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = 'pydata_sphinx_theme'
# html_theme_options = {
#     "light_css_variables": {
#         "color-brand-primary": "#2962ff",    # Strong blue for main brand
#         "color-brand-content": "#1976d2",    # Slightly darker blue for content
#         "color-highlight-on-target": "#e3f2fd",  # Very light blue for highlights
#     },
#     "dark_css_variables": {
#         "color-brand-primary": "#90caf9",    # Light blue for main brand in dark mode
#         "color-brand-content": "#64b5f6",    # Slightly darker blue for content
#         "color-highlight-on-target": "#0d47a1",  # Dark blue for highlights
#     },
# }
html_static_path = ['_static']
html_css_files = ['custom.css']


import os
import sys
sys.path.insert(0, os.path.abspath('../../src/showtens'))


import sys, pathlib, os

sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())

from src.showtens.util import gridify, import_torch, import_torchvision, _rescale_images
# None implemented yet

curpath = pathlib.Path(__file__).parent


def test_grid_pytorch():
    from PIL import Image

    torch = import_torch()
    transf = import_torchvision().transforms

    torchgrid = import_torchvision().utils.make_grid

    image = Image.open(os.path.join(curpath, "test_folder", "whale.png"))
    whale = transf.ToTensor()(image)  # (3,H,W)
    randAug = transf.RandomResizedCrop(whale.shape[-2:])

    whale_tile = torch.stack([randAug(whale) for _ in range(9)], dim=0)  # (10,3,H,W)

    grid = gridify(whale_tile, columns=5, pad_value=1.0, padding=0)  # (3,H',W')
    gridtorch = torchgrid(whale_tile, nrow=5, padding=0, pad_value=0.0)  # (3,H',W')

    assert torch.allclose(
        grid - gridtorch, torch.tensor(0.0)
    ), f"My grid does not correspond to pytorch make_grid in the padding 0 case. Max_diff = {torch.max(torch.abs(grid-gridtorch))}"


def test_grid_size():
    torch = import_torch()
    tens_size = [8, 32, 64]
    padding = [3, 5, 8]
    num_cols = [3, 5, 8]
    for sizu, paddingu, colu in zip(tens_size, padding, num_cols):
        tensor = torch.rand(10, 3, sizu, sizu)
        grid = gridify(tensor, columns=colu, max_width=sizu // 2 * colu, padding=0)

        assert grid.shape[-1] == sizu // 2 * colu, "Grid does not have the right width"

        max_width = 104
        grid = gridify(tensor, columns=colu, max_width=max_width, padding=3)

        new_W = sizu + paddingu * 2
        reshape_ratio = max_width / (new_W * colu)
        if reshape_ratio < 1:
            image_size = int((new_W) * reshape_ratio)
        else:
            image_size = new_W

        assert (
            grid.shape[-1] == image_size * colu
        ), f"Expected width {image_size*colu}, but got {grid.shape[-1]}"


def test_rescale():
    torch = import_torch()
    a = torch.randn(5, 3, 32, 32)
    reshaped_a = a.reshape(5, 3, 32 * 32)

    just_clip = _rescale_images(a, clamp_range=(0.0, 1.0), rescale=False)
    assert torch.max(just_clip) <= 1.0 and torch.min(just_clip) >= 0.0, "Rescale failed to clip to range"

    rescaled = _rescale_images(a, clamp_range=None, rescale=True)
    maxes, max_ind = torch.max(reshaped_a, dim=2, keepdim=True)  # (5,3,1) maxes for each channel
    mins, min_ind = torch.min(reshaped_a, dim=2, keepdim=True)  # (5,3,1) mins for each channel
    reshaped_rescaled_a = rescaled.reshape(5, 3, 32 * 32)

    rescaled_maxes = torch.gather(reshaped_rescaled_a, 2, max_ind)  # should all be 1
    rescaled_mins = torch.gather(reshaped_rescaled_a, 2, min_ind)  # should all be 0

    assert torch.allclose(
        rescaled_maxes, torch.tensor(1.0)
    ), "Rescale failed to rescale maxes PER CHANNELIMAGE"
    assert torch.allclose(rescaled_mins, torch.tensor(0.0)), "Rescale failed to rescale mins PER CHANNELIMAGE"

    rescaled = _rescale_images(a, clamp_range=(-1.0, 1.0), rescale=True)
    where_clamped_low = a < -1.0
    where_clamped_high = a > 1.0

    assert torch.allclose(rescaled[where_clamped_low], torch.tensor(0.0)), "Rescale failed to clamp low"
    assert torch.allclose(rescaled[where_clamped_high], torch.tensor(1.0)), "Rescale failed to clamp high"


# test_rescale()



from .imports import import_torch, import_torchvision
import os

torch = import_torch()


@torch.no_grad()
def gridify(
    tensor: torch.Tensor,
    max_width: int | None = None,
    columns: int | None = None,
    padding: int = 3,
    pad_value: float = 0.0,
) -> torch.Tensor:
    """
    Makes a grid of images/videos from a batch of images.
    Like torchvision's make_grid, but more flexible.
    Accepts (B,\\*,H,W)

    Args:
        tensor : (B,\\*,H,W) tensor
        max_width : max width of the output grid. Resizes images to fit the width
        columns : number of columns of the grid. If None, uses 8 or less
        padding : padding to add to the images
        pad_value : color of the padding

    Returns:
        (\\*,H',W') tensor, representing the grid of images/videos
    """
    transf = import_torchvision().transforms

    B, H, W = tensor.shape[0], tensor.shape[-2], tensor.shape[-1]
    device = tensor.device
    if columns is not None:
        numCol = columns
    else:
        numCol = min(8, B)

    black_cols = (-B) % numCol
    tensor = torch.cat(
        [tensor, torch.zeros(black_cols, *tensor.shape[1:], device=device)], dim=0
    )  # (B',*,H,W)
    tensor = transf.Pad(padding, fill=pad_value)(tensor)  # (B',*,H+padding*2,W+padding*2)

    B, H, W = tensor.shape[0], tensor.shape[-2], tensor.shape[-1]
    rest_dim = tensor.shape[1:-2]

    rest_dim_prod = 1
    for dim in rest_dim:
        rest_dim_prod *= dim

    if max_width is not None:
        resize_ratio = max_width / (W * numCol)
        if resize_ratio < 1:
            indiv_tens_size = int(H * resize_ratio), int(W * resize_ratio)
            tensor = tensor.reshape((B, rest_dim_prod, H, W))
            tensor = transf.Resize(indiv_tens_size, antialias=True)(tensor)  # (B',rest_dim_prod,H',W')

    B, H, W = tensor.shape[0], tensor.shape[-2], tensor.shape[-1]
    assert B % numCol == 0

    numRows = B // numCol

    tensor = tensor.reshape((numRows, numCol, rest_dim_prod, H, W))  # (numRows,numCol,rest_dim_prod,H',W')
    tensor = torch.einsum("nmrhw->rnhmw", tensor)  # (rest_prod,numRows,H',numCol,W')
    tensor = tensor.reshape((rest_dim_prod, numRows * H, numCol * W))  # (rest_prod,numRows*H,numCol*W)
    tensor = tensor.reshape((*rest_dim, numRows * H, numCol * W))  # (*,numRows*H,numCol*W)

    return tensor


@torch.no_grad()
def _create_folder(folder: str, create_folder: bool = True):
    if create_folder:
        os.makedirs(folder, exist_ok=True)
    else:
        if not (os.path.exists(folder)):
            raise FileNotFoundError(f"Folder {folder} does not exist !")


@torch.no_grad()
def _format_image(
    tensor: torch.Tensor,
    columns: int = None,
    rescale: bool = False,
    clamp_range: tuple[float] = (0.0, 1.0),
    max_width: int = None,
    padding: int = 3,
    pad_value: float = 0.0,
):
    """
    Shows tensor as an image using pyplot.
    Any extra dimensions **(\\*,C,H,W)** are treated as batch dimensions.

    Args:
        tensor : (H,W) or (C,H,W) or (\\*,C,H,W) tensor to display
        columns : number of columns to use for the grid of images (default 8 or less)
        max_width : maximum width of the image
        padding : number of pixels between images in the grid
        pad_value : inter-padding value for the grid of images
    """
    tensor = tensor.detach().cpu()

    extra_params = dict(
        columns=columns,
        max_width=max_width,
        pad_value=pad_value,
        padding=padding,
        rescale=rescale,
        clamp_range=clamp_range,
    )
    if len(tensor.shape) == 2:
        # Add batch and channel dimensions
        return _format_image(tensor[None, :, :], **extra_params)
    elif len(tensor.shape) == 3:
        # Reached (C,H,W)
        tensor = _rescale_images(tensor, rescale=rescale, clamp_range=clamp_range)
        return tensor
    elif len(tensor.shape) == 4:
        # Gridify assuming (B,C,H,W)
        B = tensor.shape[0]
        if columns is not None:
            numCol = columns
        else:
            numCol = min(8, B)

        tensor = _rescale_images(tensor, rescale=rescale, clamp_range=clamp_range)  # (B,C,H,W), rescaled
        tensor = gridify(
            tensor, columns=numCol, max_width=max_width, pad_value=pad_value, padding=padding
        )  # (C,H',W')

        return tensor
    elif len(tensor.shape) > 4:
        # Collapse extra dimension to batch
        tensor = tensor.reshape(
            (-1, tensor.shape[-3], tensor.shape[-2], tensor.shape[-1])
        )  # assume all batch dimensions
        print("Assuming extra dimension are all batch dimensions, newshape : ", tensor.shape)
        return _format_image(tensor, **extra_params)
    else:
        raise Exception(f"Tensor shape should be (H,W), (C,H,W) or (*,C,H,W), but got : {tensor.shape} !")


@torch.no_grad()
def _rescale_images(tensor: torch.Tensor, rescale: bool, clamp_range: tuple[float] = None):
    """
    Clamp to clamp_range if not None, then rescale the images to 0-1 range.
    The rescaling is dynamic with min and max values for each image channel.

    Args:
        tensor : (B,C,H,W) or (C,H,W) tensors to display
        rescale : whether to rescale the images to 0-1 range
        clamp_range : (min,max), tuple of values to clamp the images to, before rescaling
    """
    if clamp_range is not None:
        tensor = tensor.clamp(min=clamp_range[0], max=clamp_range[1])

    wasthree = False
    if len(tensor.shape) == 3:
        tensor = tensor[None, :, :, :]
        wasthree = True

    B, C, H, W = tensor.shape
    tensor = tensor.reshape(B, C, H * W)
    if rescale:
        maxes, _ = torch.max(tensor, dim=2, keepdim=True)  # (B,C,1) Max value for each image channel
        mins, _ = torch.min(tensor, dim=2, keepdim=True)
        tensor = (tensor - mins) / (torch.maximum(maxes - mins, torch.tensor(1e-5)))  # rescale
    tensor = tensor.reshape(B, C, H, W)
    if wasthree:
        tensor = tensor[0]
    return tensor
